using System;
using System.Collections.Generic;
using System.Linq;
using Mono.Cecil;
using Mono.Linker;
using Mono.Linker.Steps;
using Mono.Tuner;
using Xamarin.Android.Tasks;

namespace MonoDroid.Tuner {

	public class MarkJavaObjects :
#if ILLINK
		BaseMarkHandler
#else   // !ILLINK
		BaseSubStep
#endif  // !ILLINK
	{
		Dictionary<ModuleDefinition, Dictionary<string, TypeDefinition>> module_types = new Dictionary<ModuleDefinition, Dictionary<string, TypeDefinition>> ();

#if ILLINK
		public override void Initialize (LinkContext context, MarkContext markContext)
		{
			base.Initialize (context, markContext);
			markContext.RegisterMarkTypeAction (type => ProcessType (type));
		}
#else   // !ILLINK
		public override SubStepTargets Targets {
			get { return SubStepTargets.Type; }
		}
#endif  // !ILLINK

		public
#if !ILLINK
		override
#endif  // !ILLINK
		void ProcessType (TypeDefinition type)
		{
			// If this isn't a JLO or IJavaObject implementer,
			// then we don't need to MarkJavaObjects
			if (!type.ImplementsIJavaObject ())
				return;

			PreserveJavaObjectImplementation (type);

			if (IsImplementor (type))
				PreserveImplementor (type);

			// If a user overrode a method, we need to preserve it,
			// because it won't be referenced anywhere, but it will
			// be called from Java
			if (IsUserType (type) && type.HasMethods) {
				foreach (var method in type.Methods.Where (m => m.Overrides != null))
					PreserveMethod (type, method);
			}
		}

		void PreserveJavaObjectImplementation (TypeDefinition type)
		{
			PreserveIntPtrConstructor (type);
			PreserveAttributeSetConstructor (type);
			PreserveInvoker (type);
#if ILLINK
			PreserveInterfaces (type);
#endif // ILLINK
		}

		void PreserveAttributeSetConstructor (TypeDefinition type)
		{
			if (!type.HasMethods)
				return;

			foreach (var constructor in GetAttributeSetConstructors (type))
				PreserveMethod (type, constructor);
		}

		static IEnumerable<MethodDefinition> GetAttributeSetConstructors (TypeDefinition type)
		{
			foreach (MethodDefinition constructor in type.Methods.Where (m => m.IsConstructor)) {
				if (!constructor.HasParameters)
					continue;

				var parameters = constructor.Parameters;

				if (parameters.Count < 2 || parameters.Count > 3)
					continue;

				if (parameters [0].ParameterType.FullName != "Android.Content.Context")
					continue;

				if (parameters [1].ParameterType.FullName != "Android.Util.IAttributeSet")
					continue;

				if (parameters.Count == 3 && parameters [2].ParameterType.FullName != "System.Int32")
					continue;

				yield return constructor;
			}
		}

		void PreserveIntPtrConstructor (TypeDefinition type)
		{
			var constructor = GetIntPtrConstructor (type);

			if (constructor != null)
				PreserveMethod (type, constructor);

			var constructor2 = GetNewIntPtrConstructor (type);

			if (constructor2 != null)
				PreserveMethod (type, constructor2);
		}

		static MethodDefinition GetIntPtrConstructor (TypeDefinition type)
		{
			if (!type.HasMethods)
				return null;

			foreach (MethodDefinition constructor in type.Methods.Where (m => m.IsConstructor)) {
				if (!constructor.HasParameters)
					continue;

				if (constructor.Parameters.Count != 1 || constructor.Parameters[0].ParameterType.FullName != "System.IntPtr")
					continue;

				return constructor;
			}

			return null;
		}

		static MethodDefinition GetNewIntPtrConstructor (TypeDefinition type)
		{
			if (!type.HasMethods)
				return null;

			foreach (MethodDefinition constructor in type.Methods.Where (m => m.IsConstructor)) {
				if (!constructor.HasParameters)
					continue;

				if (constructor.Parameters.Count != 2 || constructor.Parameters[0].ParameterType.FullName != "System.IntPtr"
					|| constructor.Parameters[1].ParameterType.FullName != "Android.Runtime.JniHandleOwnership")
					continue;

				return constructor;
			}

			return null;
		}

		void PreserveMethod (TypeDefinition type, MethodDefinition method)
		{
			Annotations.AddPreservedMethod (type, method);
		}

		string TypeNameWithoutKey (string name)
		{
			var idx = name.IndexOf (", PublicKeyToken=", StringComparison.Ordinal);
			if (idx > 0)
				name = name.Substring (0, idx);

			return name;
		}

		bool CheckInvokerType (TypeDefinition type, string name)
		{
			return TypeNameWithoutKey (name) == TypeNameWithoutKey ($"{ type.FullName}, { type.Module.Assembly.FullName}");
		}

		void PreserveInterfaceMethods (TypeDefinition type, TypeDefinition invoker)
		{
			foreach (var m in type.Methods.Where (m => !m.IsConstructor)) {
				string methodAndType;
				if (!m.TryGetRegisterMember (out methodAndType))
					continue;

				if (!methodAndType.Contains (":"))
					continue;

				var values = methodAndType.Split (new char [] { ':' }, 2);
				if (!CheckInvokerType (invoker, values [1]))
					continue;

				foreach (var invokerMethod in invoker.Methods.Where (m => !m.IsConstructor)) {
					if (invokerMethod.Name == values [0]) {
						PreserveMethod (invoker, invokerMethod);
						break;
					}
				}
			}
		}

		void PreserveInvoker (TypeDefinition type)
		{
			var invoker = GetInvokerType (type);
			if (invoker == null)
				return;

			PreserveConstructors (type, invoker);
			PreserveIntPtrConstructor (invoker);
			PreserveInterfaceMethods (type, invoker);
#if ILLINK
			PreserveInterfaces (invoker);
#endif // ILLINK
		}

#if ILLINK
		void PreserveInterfaces (TypeDefinition type)
		{
			if (!type.HasInterfaces)
				return;

			if (!ShouldPreserveInterfaces (type))
				return;

			foreach (var iface in type.Interfaces) {
				var td = iface.InterfaceType.Resolve ();
				if (!td.ImplementsIJavaPeerable ())
					continue;
				Annotations.Mark (td);
			}
		}

		// False if [Register(DoNotGenerateAcw=true)] is on the type
		// False if [JniTypeSignature(GenerateJavaPeer=false)] is on the type
		bool ShouldPreserveInterfaces (TypeDefinition type)
		{
			if (!type.HasCustomAttributes)
				return true;

			foreach (var attr in type.CustomAttributes) {
				switch (attr.AttributeType.FullName) {
					case "Android.Runtime.RegisterAttribute":
						foreach (var property in attr.Properties) {
							if (property.Name == "DoNotGenerateAcw") {
								if (property.Argument.Value is bool value && value)
									return false;
								break;
							}
						}
						break;
					case "Java.Interop.JniTypeSignatureAttribute":
						foreach (var property in attr.Properties) {
							if (property.Name == "GenerateJavaPeer") {
								if (property.Argument.Value is bool value && !value)
									return false;
								break;
							}
						}
						break;
					default:
						break;
				}
			}

			return true;
		}
#endif // ILLINK

		TypeDefinition GetInvokerType (TypeDefinition type)
		{
			const string suffix = "Invoker";
			string fullname = type.FullName;

			if (type.HasGenericParameters) {
				var pos = fullname.IndexOf ('`');
				if (pos == -1)
					throw new ArgumentException ();

				fullname = fullname.Substring (0, pos) + suffix + fullname.Substring (pos);
			} else
				fullname = fullname + suffix;

			return FindType (type, fullname);
		}

		// Keep a dictionary cache of all types in a module rather than
		// looping through them on every lookup.
		TypeDefinition FindType (TypeDefinition type, string fullname)
		{
			Dictionary<string, TypeDefinition> types;

			if (!module_types.TryGetValue (type.Module, out types)) {
				types = GetTypesInModule (type.Module);
				module_types.Add (type.Module, types);
			}

			TypeDefinition helper;

			if (types.TryGetValue (fullname, out helper))
				return helper;

			return null;
		}

		static Dictionary<string, TypeDefinition> GetTypesInModule (ModuleDefinition module)
		{
			var types = module.Types.ToDictionary (p => p.FullName);

			foreach (var t in module.Types)
				AddNestedTypes (types, t);

			return types;
		}

		static void AddNestedTypes (Dictionary<string, TypeDefinition> types, TypeDefinition type)
		{
			if (!type.HasNestedTypes)
				return;

			foreach (var t in type.NestedTypes) {
				types.Add (t.FullName, t);
				AddNestedTypes (types, t);
			}
		}

		void PreserveConstructors (TypeDefinition type, TypeDefinition helper)
		{
			if (!helper.HasMethods)
				return;

			foreach (MethodDefinition ctor in helper.Methods.Where (m => m.IsConstructor))
				PreserveMethod (type, ctor);
		}

		static bool IsImplementor (TypeDefinition type)
		{
			return type.Name.EndsWith ("Implementor", StringComparison.Ordinal) && type.Inherits ("Java.Lang.Object");
		}

		static bool IsUserType (TypeDefinition type)
		{
			return !MonoAndroidHelper.IsFrameworkAssembly (type.Module.Assembly.Name.Name + ".dll");
		}

		void PreserveImplementor (TypeDefinition type)
		{
			if (!type.HasMethods)
				return;

			foreach (MethodDefinition method in type.Methods)
				if (method.Name.EndsWith ("Handler", StringComparison.Ordinal))
					PreserveMethod (type, method);
		}
	}
}
